
{*******************************************************}
{                                                       }
{       Turbo Pascal Version 7.0                        }
{       Multi-Sprites Editor Unit                       }
{                                                       }
{       Copyright (c) 1994,95 by Solar Designer         }
{                                                       }
{*******************************************************}

unit LPalette;
interface
uses
   Objects, Memory,
   Tables,
   GApp, GDrivers, GPalette,
   SEditor, ReadFmt;

const
   cmPaletteLoad = $E0;
   cmPaletteSave = $E1;
   cmPaletteStd =  $E2;

   cmShowUsed =    $E8;

var
   StandardPalette :Array [0..15, 0..2] of Byte;

procedure LoadPalette(var S            :TStream;
                      PalFmt           :TPaletteFormat);

procedure SavePalette(var S            :TStream);

implementation

function  GetNearestColor(var RGB      :Array of Byte;
                          var Palette  :TRGBPalette)       :Byte;

function  Diff(c1, c2                  :Integer) :LongInt;
begin
   Diff:=Sqr(LongInt(c1-c2));
end;

var
   i            :Integer;
   d, dmin      :LongInt;

begin
   GetNearestColor:=0; dmin:=$FFFF;
   for i:=0 to 255 do
   begin
      d:=Diff(RGB[0], Palette[i, 0]) +
         Diff(RGB[1], Palette[i, 1]) +
         Diff(RGB[2], Palette[i, 2]);
      if d<dmin then
      begin
         dmin:=d; GetNearestColor:=i;
      end;
   end;
end;

procedure LoadPalette;
type
   TWindowsColor =
   record
      B, G, R, C   :Byte;
   end;
var
   WinP            :Array [0..255] of TWindowsColor;
   P               :TRGBPalette absolute WinP;
   WinC            :TWindowsColor;
   C, NC           :Integer;
   RN              :Boolean;
begin
   if PalFmt=pfPCX then
   begin
      S.Read(P, 1);
      if P[0, 0]<>12 then Exit;
   end;

   if PalFmt=pfWindows
   then S.Read(WinP, SizeOf(WinP))
   else S.Read(P, SizeOf(P));
   if S.Status<>stOk then
   begin
      GDrivers.ShowPalette(P); Exit;
   end;

   for C:=0 to 255 do
   case PalFmt of
      pfWindows:
      begin
         LongInt(WinC):=LongInt(WinP[C]);
         P[C, 0]:=WinC.R shr 2; P[C, 1]:=WinC.G shr 2; P[C, 2]:=WinC.B shr 2;
      end;
      pfPCX:
      begin
         P[C, 0]:=P[C, 0] shr 2; P[C, 1]:=P[C, 1] shr 2; P[C, 2]:=P[C, 2] shr 2;
      end;
   end;

   RN:=False;
   for C:=0 to 15 do
   begin
      NC:=GetNearestColor(StandardPalette[C], P);
      if ColorTable[C]<>NC then RN:=True;
      ColorTable[C]:=NC;
   end;

   Application^.EnableCommands([cmPaletteStd]);
   if RN then
   begin
      GDrivers.HidePalette(PaletteRGB);
      PaletteRGB:=P;
      Application^.Redraw;
      GDrivers.ShowPalette(P);
   end else
   begin
      PaletteRGB:=P;
      GDrivers.SetPalette(P);
      InfoBar^.DrawView;
   end;

   if cRGB2C<>nil then DisposeCache(cRGB2C);
end;

procedure SavePalette;
begin
   S.Write(PaletteRGB, SizeOf(PaletteRGB));
end;

end.
