
{*******************************************************}
{                                                       }
{       Turbo Pascal Version 7.0                        }
{       Multi-Sprites Editor Unit                       }
{                                                       }
{       Copyright (c) 1994,95 by Solar Designer         }
{                                                       }
{*******************************************************}

unit ReadFmt;
interface
{$I-}
uses
   Objects, GDrivers, GraphDrv,
   HistList;

type
   TPaletteFormat =
   (pfRaw, pfWindows, pfPCX);

   TImageFormat =
   record
      Name,
      FormatId     :PString;
      HeaderSize,
      FSizeOfs,
      FSizeSize,
      SizesOfs,
      SizesSize,
      FSizeInc,
      SizesInc,
      PaletteOfs   :Integer;
      PaletteFmt   :TPaletteFormat;
      Reverse      :Boolean;
      Packing      :(ipNone, ipPCX);
      Multiple     :Boolean;
   end;

const
   MaxFormats =    32;
   FormatsFile =   'MSEDIT.FMT';

   hiImage =       100;

var
   Formats         :Array [0..MaxFormats] of TImageFormat;
const
   FormatsCount    :Integer= 0;
   MaxHeaderSize   :Integer= 0;

procedure ReadFormats;

implementation

procedure ReadFormats;
var
   f               :Text;
   S               :String[80];
   Op              :String[23];
   Ln, p, i        :Integer;
   LnS             :String[7];
   PS              :PString;

function  Value(Min, Max               :Integer) :Integer;
var
   Res, Error      :Integer;
begin
   Val(Copy(S, p+1, 255), Res, Error);
   if Error<>0 then
      Abort('Invalid number in '+FormatsFile+', line #'+LnS);
   if (Res<Min) or (Res>Max) then
      Abort('Number not in valid range in '+FormatsFile+', line #'+LnS);
   Value:=Res;
end;

function  StrParam                               :PString;
var
   i               :Integer;

function  GetHex(c                     :Char)    :Integer;
begin
   c:=UpCase(c);
   if (c>='0') and (c<='9') then GetHex:=Ord(c)-Ord('0') else
   if (c>='A') and (c<='F') then GetHex:=Ord(c)-(Ord('A')-10) else GetHex:=-1;
end;

begin
   Delete(S, 1, p);
   if (S[0]<=#2) or (S[1]<>'''') or (S[Length(S)]<>'''') then
      Abort('Invalid string parameter in '+FormatsFile+', line #'+LnS);
   Delete(S, 1, 1); Dec(S[0]);

   i:=1;
   repeat
      if (S[i]='\') and (GetHex(S[i+1])>=0) and (GetHex(S[i+2])>=0) then
      begin
         S[i]:=Chr(GetHex(S[i+1]) shl 4 + GetHex(S[i+2]));
         Delete(S, i+1, 2);
      end;
      Inc(i);
   until i>=Length(S);

   StrParam:=NewStr(S);
end;

begin
   FillChar(Formats, SizeOf(Formats), $FF);
   for i:=1 to MaxFormats do
   with Formats[i] do
   begin
      FSizeInc:=0; SizesInc:=0; PaletteOfs:=-1; PaletteFmt:=pfRaw;
      FormatId:=nil; Multiple:=False; Reverse:=False; Packing:=ipNone;
   end;

   Assign(f, GetPath+FormatsFile); Reset(f);
   Ln:=0;
   while not EOF(f) do
   begin
      Inc(Ln); Str(Ln, LnS);
      ReadLn(f, S);

      while (S<>'') and ((S[1]=' ') or (S[1]=#9)) do Delete(S, 1, 1);
      while (S<>'') and ((S[Length(S)]=' ') or (S[Length(S)]=#9)) do Dec(S[0]);

      if (S='') or (S[1]=';') then Continue;

      p:=Pos('=', S);
      Op:=S; if p>0 then Op[0]:=Chr(p-1);
      for i:=1 to Length(Op) do Op[i]:=UpCase(Op[i]);

      if Op='FORMAT' then
      begin
         Inc(FormatsCount);
         if FormatsCount>MaxFormats then
            Abort('Too many format definitions in '+FormatsFile);
         Formats[FormatsCount].Name:=StrParam;
      end else
      with Formats[FormatsCount] do
      if Op='FORMATID' then FormatId:=StrParam else
      if Op='HEADERSIZE' then HeaderSize:=Value(4, 2048) else
      if Op='FILESIZEOFFSET' then FSizeOfs:=Value(0, 2048) else
      if Op='SIZESOFFSET' then SizesOfs:=Value(0, 2048) else
      if Op='FILESIZEINCREMENT' then FSizeInc:=Value(-2048, 2048) else
      if Op='SIZESINCREMENT' then SizesInc:=Value(-1024, 1024) else
      if Op='FILESIZEFIELDSIZE' then FSizeSize:=Value(2, 4) else
      if Op='SIZESFIELDSIZE' then SizesSize:=Value(2, 4) else
      if Op='PALETTEOFFSET' then PaletteOfs:=Value(-2048, 2048) else
      if Op='PALETTEFORMAT' then
      begin
         PS:=StrParam;
         for i:=1 to Length(PS^) do PS^[i]:=UpCase(PS^[i]);
         if PS^='RAW' then PaletteFmt:=pfRAW else
         if PS^='WINDOWS' then PaletteFmt:=pfWindows else
         if PS^='PCX' then PaletteFmt:=pfPCX else
            Abort('Invalid palette format in '+FormatsFile+', line #'+LnS);
         DisposeStr(PS);
      end else
      if Op='MULTIPLEIMAGES' then Multiple:=True else
      if Op='REVERSE' then Reverse:=True else
      if Op='PACKING' then
      begin
         PS:=StrParam;
         for i:=1 to Length(PS^) do PS^[i]:=UpCase(PS^[i]);
         if PS^='NONE' then Packing:=ipNone else
         if PS^='PCX' then Packing:=ipPCX else
            Abort('Invalid packing method in '+FormatsFile+', line #'+LnS);
         DisposeStr(PS);
      end else
      if Op='EXTENSION' then
      begin
         PS:=StrParam;
         if PS^[0]>#3 then
            Abort('Invalid extension in '+FormatsFile+', line #'+LnS);
         HistoryAdd(hiImage, '*.'+PS^); DisposeStr(PS);
      end
      else Abort('Invalid command in '+FormatsFile+', line #'+LnS);
   end;
   Close(f);
   if IOResult<>0 then Abort('Unable to read '+FormatsFile);
   if FormatsCount=0 then Abort('No formats defined in '+FormatsFile);

   for i:=1 to FormatsCount do
   with Formats[i] do
   begin
      if (HeaderSize<0) or (SizesOfs<0) or
         ((FSizeOfs>=0) and (FSizeSize<0)) or (SizesSize<0) then
         Abort('Not enough parameters for image format '''+Name^+''' in '+FormatsFile);
      if MaxHeaderSize<HeaderSize then MaxHeaderSize:=HeaderSize;
   end;
end;

end.
